# pylint: disable=missing-docstring,expression-not-assigned,unused-variable

import os

import pytest
from expecter import expect

from yorm import diskutils


def describe_touch():

    @pytest.fixture
    def new_path(tmpdir):
        tmpdir.chdir()
        return os.path.join('.', 'file.ext')

    @pytest.fixture
    def new_path_in_directory():
        dirpath = os.path.join('path', 'to', 'directory')
        return os.path.join(dirpath, 'file.ext')

    def it_creates_files(new_path):
        diskutils.touch(new_path)
        expect(os.path.exists(new_path)).is_true()

    def it_can_be_called_twice(new_path):
        diskutils.touch(new_path)
        diskutils.touch(new_path)
        expect(os.path.exists(new_path)).is_true()

    def it_creates_missing_directories(new_path_in_directory):
        diskutils.touch(new_path_in_directory)
        expect(os.path.exists(new_path_in_directory)).is_true()


def describe_delete():

    @pytest.fixture
    def existing_path(tmpdir):
        tmpdir.chdir()
        path = "tmp/path/to/file.ext"
        os.makedirs(os.path.dirname(path))
        open(path, 'w').close()
        return path

    @pytest.fixture
    def existing_dirpath(tmpdir):
        tmpdir.chdir()
        dirpath = "tmp/path/to/directory"
        os.makedirs(dirpath)
        return dirpath

    def it_deletes_existing_files(existing_path):
        diskutils.delete(existing_path)
        expect(os.path.exists(existing_path)).is_false()

    def it_ignores_missing_files():
        diskutils.delete("tmp/path/to/non/file")

    def it_deletes_directories(existing_dirpath):
        diskutils.delete(existing_dirpath)
        expect(os.path.exists(existing_dirpath)).is_false()



# pylint: disable=unused-variable,redefined-outer-name,expression-not-assigned,singleton-comparison

import logging
from unittest.mock import Mock

import pytest
from expecter import expect

import yorm
from yorm import exceptions
from yorm import utilities

log = logging.getLogger(__name__)


@pytest.fixture
def model_class(tmpdir):
    tmpdir.chdir()

    @yorm.sync("data/{self.kind}/{self.key}.yml", auto_create=False)
    class Model:

        def __init__(self, kind, key, **kwargs):
            self.kind = kind
            self.key = key
            assert 0 <= len(kwargs) < 2
            if kwargs:
                assert kwargs == {'test': 'test'}

        def __eq__(self, other):
            return (self.kind, self.key) == (other.kind, other.key)

    return Model


@pytest.fixture
def instance(model_class):
    return model_class('foo', 'bar')


@pytest.fixture
def instances(model_class):
    instances = [
        model_class(kind, key)
        for kind in ('spam', 'egg')
        for key in ('foo', 'bar')
    ]
    for instance in instances:
        instance.__mapper__.create()

    return instances


def describe_create():

    def it_creates_files(model_class):
        instance = utilities.create(model_class, 'foo', 'bar')

        expect(instance.__mapper__.exists) == True
        expect(instance.__mapper__.modified) == False

    def it_requires_files_to_not_yet_exist(model_class, instance):
        instance.__mapper__.create()

        with expect.raises(exceptions.DuplicateMappingError):
            utilities.create(model_class, 'foo', 'bar')

    def it_can_overwrite_files(model_class, instance):
        instance.__mapper__.create()

        utilities.create(model_class, 'foo', 'bar', overwrite=True)

    def it_supports_keyword_arguments(model_class):
        instance = utilities.create(model_class, 'foo', key='bar')

        expect(instance.__mapper__.exists) == True

    def it_can_also_be_called_with_an_instance(instance):
        expect(yorm.create(instance)) == instance

    def it_requires_a_mapped_class_or_instance():
        with expect.raises(TypeError):
            utilities.create(Mock)


def describe_find():

    def it_returns_object_when_found(model_class, instance):
        instance.__mapper__.create()

        expect(utilities.find(model_class, 'foo', 'bar')) == instance

    def it_returns_none_when_no_match(model_class):
        expect(utilities.find(model_class, 'not', 'here')) == None

    def it_allows_objects_to_be_created(model_class):
        expect(utilities.find(model_class, 'new', 'one', create=True)) == \
            model_class('new', 'one')

    def it_supports_keyword_arguments(model_class, instance):
        instance.__mapper__.create()

        expect(utilities.find(model_class, 'foo', key='bar')) == instance

    def it_can_also_be_called_with_an_instance(instance):
        expect(yorm.find(instance, create=True)) == instance

    def it_requires_a_mapped_class_or_instance():
        with expect.raises(TypeError):
            utilities.find(Mock)


def describe_match():

    def with_class_and_factory(model_class, instances):
        matches = list(
            utilities.match(
                model_class,
                (lambda key, kind: model_class(kind, key, test="test")),
                kind='spam',
                key='foo',
            )
        )
        expect(len(matches)) == 1
        instance = matches[0]
        expect(instance.kind) == 'spam'
        expect(instance.key) == 'foo'
        expect(instances).contains(instance)

    def with_class_and_no_factory(model_class, instances):
        matches = list(
            utilities.match(
                model_class,
                kind='spam',
                key='foo',
            )
        )
        expect(len(matches)) == 1
        instance = matches[0]
        expect(instance.kind) == 'spam'
        expect(instance.key) == 'foo'
        expect(instances).contains(instance)

    def with_string(model_class, instances):
        matches = list(
            utilities.match(
                "data/{kind}/{key}.yml",
                (lambda key, kind: model_class(kind, key, test="test")),
                kind='egg',
                key='foo',
            )
        )
        expect(len(matches)) == 1
        instance = matches[0]
        expect(instance.kind) == 'egg'
        expect(instance.key) == 'foo'
        expect(instances).contains(instance)

    def with_self_string(model_class, instances):
        matches = list(
            utilities.match(
                "data/{self.kind}/{self.key}.yml",
                (lambda key, kind: model_class(kind, key, test="test")),
                kind='spam',
                key='bar',
            )
        )
        expect(len(matches)) == 1
        instance = matches[0]
        expect(instance.kind) == 'spam'
        expect(instance.key) == 'bar'
        expect(instances).contains(instance)

    def with_class_and_partial_match(model_class, instances):
        matches = list(
            utilities.match(
                model_class,
                kind='spam',
            )
        )
        expect(len(matches)) == 2
        for instance in matches:
            expect(instance.kind) == 'spam'
            expect(instances).contains(instance)

    def with_string_and_partial_match(model_class, instances):
        matches = list(
            utilities.match(
                "data/{kind}/{key}.yml",
                (lambda key, kind: model_class(kind, key, test="test")),
                key='foo',
            )
        )
        expect(len(matches)) == 2
        for instance in matches:
            expect(instance.key) == 'foo'
            expect(instances).contains(instance)

    def with_self_string_and_partial_match(model_class, instances):
        matches = list(
            utilities.match(
                "data/{self.kind}/{self.key}.yml",
                (lambda key, kind: model_class(kind, key, test="test")),
                kind='egg',
            )
        )
        expect(len(matches)) == 2
        for instance in matches:
            expect(instance.kind) == 'egg'
            expect(instances).contains(instance)


def describe_load():

    def it_marks_files_as_unmodified(instance):
        instance.__mapper__.create()
        instance.__mapper__.modified = True

        utilities.load(instance)

        expect(instance.__mapper__.modified) == False

    def it_requires_a_mapped_instance():
        with expect.raises(TypeError):
            utilities.load(Mock)


def describe_save():

    def it_creates_files(instance):
        utilities.save(instance)

        expect(instance.__mapper__.exists) == True

    def it_marks_files_as_modified(instance):
        instance.__mapper__.create()
        instance.__mapper__.modified = False

        utilities.save(instance)

        expect(instance.__mapper__.modified) == True

    def it_expects_the_file_to_not_be_deleted(instance):
        instance.__mapper__.delete()

        with expect.raises(exceptions.DeletedFileError):
            utilities.save(instance)

    def it_requires_a_mapped_instance():
        with expect.raises(TypeError):
            utilities.save(Mock)


def describe_delete():

    def it_deletes_files(instance):
        utilities.delete(instance)

        expect(instance.__mapper__.exists) == False
        expect(instance.__mapper__.deleted) == True

    def it_requires_a_mapped_instance():
        with expect.raises(TypeError):
            utilities.delete(Mock)



# pylint: disable=missing-docstring,no-self-use,abstract-class-instantiated

import pytest

from yorm.bases import Converter


class TestConverter:

    """Unit tests for the `Converter` class."""

    def test_converter_class_cannot_be_instantiated(self):
        with pytest.raises(TypeError):
            Converter()

    def test_converter_class_methods_cannot_be_called(self):
        with pytest.raises(NotImplementedError):
            Converter.create_default()
        with pytest.raises(NotImplementedError):
            Converter.to_value(None)
        with pytest.raises(NotImplementedError):
            Converter.to_data(None)



# pylint: disable=unused-variable,unused-argument,expression-not-assigned
# pylint: disable=missing-docstring,no-self-use,no-member,misplaced-comparison-constant

import logging
from unittest.mock import patch, Mock

import pytest
from expecter import expect

from yorm import decorators
from yorm.bases import Converter

log = logging.getLogger(__name__)


class MockConverter(Converter):
    """Sample converter class."""

    @classmethod
    def create_default(cls):
        return None

    @classmethod
    def to_value(cls, data):
        return None

    @classmethod
    def to_data(cls, value):
        return None


def describe_sync():

    def describe_object():

        @pytest.fixture
        def instance():
            cls = type('Sample', (), {})
            instance = cls()
            return instance

        @pytest.fixture
        def path(tmpdir):
            tmpdir.chdir()
            return "sample.yml"

        def with_no_attrs(instance, path):
            sample = decorators.sync(instance, path)

            expect(sample.__mapper__.path) == "sample.yml"
            expect(sample.__mapper__.attrs) == {}

        def with_attrs(instance, path):
            attrs = {'var1': MockConverter}
            sample = decorators.sync(instance, path, attrs)

            expect(sample.__mapper__.path) == "sample.yml"
            expect(sample.__mapper__.attrs) == {'var1': MockConverter}

        def cannot_be_called_twice(instance, path):
            sample = decorators.sync(instance, path)

            with pytest.raises(TypeError):
                decorators.sync(instance, path)

        @patch('yorm.diskutils.exists', Mock(return_value=True))
        @patch('yorm.diskutils.read', Mock(return_value="abc: 123"))
        @patch('yorm.diskutils.stamp', Mock())
        def reads_existing_files(instance, path):
            sample = decorators.sync(instance, path, auto_track=True)

            expect(sample.abc) == 123


@patch('yorm.diskutils.write', Mock())
@patch('yorm.diskutils.stamp', Mock())
@patch('yorm.diskutils.read', Mock(return_value=""))
class TestSyncInstances:
    """Unit tests for the `sync_instances` decorator."""

    @decorators.sync("sample.yml")
    class SampleDecorated:
        """Sample decorated class."""

        def __repr__(self):
            return "<decorated {}>".format(id(self))

    @decorators.sync("sample.yml", auto_track=True)
    class SampleDecoratedAutoTrack:
        """Sample decorated class with automatic attribute tracking."""

        def __repr__(self):
            return "<decorated {}>".format(id(self))

    @decorators.sync("{UUID}.yml")
    class SampleDecoratedIdentifiers:
        """Sample decorated class using UUIDs for paths."""

        def __repr__(self):
            return "<decorated w/ UUID {}>".format(id(self))

    @decorators.sync("tmp/path/to/{n}.yml", {'n': 'name'})
    class SampleDecoratedAttributes:
        """Sample decorated class using an attribute value for paths."""

        def __init__(self, name):
            self.name = name

        def __repr__(self):
            return "<decorated w/ specified attributes {}>".format(id(self))

    @decorators.sync("tmp/path/to/{self.name}.yml")
    class SampleDecoratedAttributesAutomatic:
        """Sample decorated class using an attribute value for paths."""

        def __init__(self, name):
            self.name = name

        def __repr__(self):
            return "<decorated w/ automatic attributes {}>".format(id(self))

    @decorators.sync("{self.a}/{self.b}/{c}.yml", {'self.b': 'b', 'c': 'c'})
    class SampleDecoratedAttributesCombination:
        """Sample decorated class using an attribute value for paths."""

        def __init__(self, a, b, c):
            self.a = a
            self.b = b
            self.c = c

        def __repr__(self):
            return "<decorated w/ attributes {}>".format(id(self))

    @decorators.sync("sample.yml", attrs={'var1': MockConverter})
    class SampleDecoratedWithAttributes:
        """Sample decorated class using a single path."""

    def test_no_attrs(self):
        """Verify mapping can be enabled with no attributes."""
        sample = self.SampleDecorated()

        expect(sample.__mapper__.path) == "sample.yml"
        expect(sample.__mapper__.attrs) == {}

    def test_with_attrs(self):
        """Verify mapping can be enabled with with attributes."""
        sample = self.SampleDecoratedWithAttributes()
        assert "sample.yml" == sample.__mapper__.path
        assert ['var1'] == list(sample.__mapper__.attrs.keys())

    @patch('yorm.diskutils.exists', Mock(return_value=True))
    def test_init_existing(self):
        """Verify an existing file is read."""
        with patch('yorm.diskutils.read', Mock(return_value="abc: 123")):
            sample = self.SampleDecoratedAutoTrack()
        assert 123 == sample.abc

    @patch('uuid.uuid4', Mock(return_value=Mock(hex='abc123')))
    def test_filename_uuid(self):
        """Verify UUIDs can be used for filename."""
        sample = self.SampleDecoratedIdentifiers()
        assert "abc123.yml" == sample.__mapper__.path
        assert {} == sample.__mapper__.attrs

    def test_filename_attributes(self):
        """Verify attributes can be used to determine filename."""
        sample1 = self.SampleDecoratedAttributes('one')
        sample2 = self.SampleDecoratedAttributes('two')
        assert "tmp/path/to/one.yml" == sample1.__mapper__.path
        assert "tmp/path/to/two.yml" == sample2.__mapper__.path

    def test_filename_attributes_automatic(self):
        """Verify attributes can be used to determine filename (auto save)."""
        sample1 = self.SampleDecoratedAttributesAutomatic('one')
        sample2 = self.SampleDecoratedAttributesAutomatic('two')
        assert "tmp/path/to/one.yml" == sample1.__mapper__.path
        assert "tmp/path/to/two.yml" == sample2.__mapper__.path

    def test_filename_attributes_combination(self):
        """Verify attributes can be used to determine filename (combo)."""
        log.info("Creating first object...")
        sample1 = self.SampleDecoratedAttributesCombination('A', 'B', 'C')
        log.info("Creating second object...")
        sample2 = self.SampleDecoratedAttributesCombination(1, 2, 3)
        assert "A/B/C.yml" == sample1.__mapper__.path
        assert "1/2/3.yml" == sample2.__mapper__.path


def describe_attr():

    class MockConverter1(MockConverter):
        """Sample converter class."""

    class MockConverter2(MockConverter):
        """Sample converter class."""

    @pytest.fixture
    def path(tmpdir):
        tmpdir.chdir()
        return "mock/path"

    def it_accepts_one_argument(path):

        @decorators.attr(var1=MockConverter1)
        @decorators.sync(path)
        class SampleDecoratedSingle:
            """Class using single `attr` decorator."""

        sample = SampleDecoratedSingle()
        expect(sample.__mapper__.attrs) == {'var1': MockConverter1}

    def it_rejects_zero_arguments():
        with expect.raises(ValueError):
            decorators.attr()

    def it_rejects_more_than_one_argument():
        with expect.raises(ValueError):
            decorators.attr(foo=1, bar=2)

    def it_can_be_applied_multiple_times(path):

        @decorators.attr(var1=MockConverter1)
        @decorators.attr(var2=MockConverter2)
        @decorators.sync(path)
        class SampleDecoratedMultiple:
            """Class using multiple `attr` decorators."""

        sample = SampleDecoratedMultiple()
        expect(sample.__mapper__.attrs) == {'var1': MockConverter1,
                                            'var2': MockConverter2}

    def it_can_be_applied_before_sync(path):

        @decorators.attr(var2=MockConverter2)
        @decorators.sync(path, attrs={'var1': MockConverter1})
        class SampleDecoratedCombo:
            """Class using `attr` decorator and providing a mapping."""

        sample = SampleDecoratedCombo()
        expect(sample.__mapper__.attrs) == {'var1': MockConverter1,
                                            'var2': MockConverter2}

    def it_can_be_applied_after_sync(path):

        @decorators.sync(path, attrs={'var1': MockConverter1})
        @decorators.attr(var2=MockConverter2)
        class SampleDecoratedBackwards:
            """Class using `attr` decorator after `sync` decorator."""

        sample = SampleDecoratedBackwards()
        expect(sample.__mapper__.attrs) == {'var1': MockConverter1,
                                            'var2': MockConverter2}



"""Shared internal classes and functions."""

import collections
import logging


# CONSTANTS ####################################################################

MAPPER = '__mapper__'
ALL = 'all'
UUID = 'UUID'

PRINT_VERBOSITY = 0  # minimum verbosity to using `print`
STR_VERBOSITY = 3  # minimum verbosity to use verbose `__str__`
MAX_VERBOSITY = 4  # maximum verbosity level implemented

OVERRIDE_MESSAGE = "Method must be implemented in subclasses"


# GLOBALS ######################################################################


verbosity = 0  # global verbosity setting for controlling string formatting

attrs = collections.defaultdict(collections.OrderedDict)
path_formats = {}


# LOGGING ######################################################################


logging.addLevelName(logging.DEBUG - 1, 'TRACE')


def _trace(self, message, *args, **kwargs):
    if self.isEnabledFor(logging.DEBUG - 1):
        # pylint: disable=protected-access
        self._log(logging.DEBUG - 1, message, args, **kwargs)


logging.Logger.trace = _trace


# DECORATORS ###################################################################


class classproperty(object):
    """Read-only class property decorator."""

    def __init__(self, getter):
        self.getter = getter

    def __get__(self, instance, owner):
        return self.getter(owner)


# FUNCTIONS ####################################################################


def get_mapper(obj, *, expected=None):
    """Get the `Mapper` instance attached to an object."""
    try:
        mapper = object.__getattribute__(obj, MAPPER)
    except AttributeError:
        mapper = None

    if mapper and expected is False:
        msg = "{!r} is already mapped".format(obj)
        raise TypeError(msg)

    if not mapper and expected is True:
        msg = "{!r} is not mapped".format(obj)
        raise TypeError(msg)

    return mapper


def set_mapper(obj, mapper):
    """Attach a `Mapper` instance to an object."""
    setattr(obj, MAPPER, mapper)
    return mapper



"""Custom exceptions."""

from abc import ABCMeta

import yaml


class Error(Exception, metaclass=ABCMeta):
    """Base class for all YORM exceptions."""


class DuplicateMappingError(Error, FileExistsError):
    """The file is already in use by another mapping."""


class MissingFileError(Error, FileNotFoundError):
    """An object's file has not yet been created."""


class DeletedFileError(Error, FileNotFoundError):
    """An object's file was deleted."""


class FileContentError(Error, yaml.error.YAMLError, ValueError):
    """Text could not be parsed as valid YAML."""



"""Base classes for mapping."""

import abc
import functools
import logging

from .. import common

log = logging.getLogger(__name__)


def load_before(method):
    """Decorate methods that should load before call."""

    if getattr(method, '_load_before', False):
        return method

    @functools.wraps(method)
    def wrapped(self, *args, **kwargs):
        __tracebackhide__ = True  # pylint: disable=unused-variable

        if not _private_call(method, args):
            mapper = common.get_mapper(self)
            if mapper and mapper.modified:
                log.debug("Loading before call: %s", method.__name__)
                mapper.load()
                if mapper.auto_save_after_load:
                    mapper.save()
                    mapper.modified = False

        return method(self, *args, **kwargs)

    setattr(wrapped, '_load_before', True)

    return wrapped


def save_after(method):
    """Decorate methods that should save after call."""

    if getattr(method, '_save_after', False):
        return method

    @functools.wraps(method)
    def wrapped(self, *args, **kwargs):
        __tracebackhide__ = True  # pylint: disable=unused-variable

        result = method(self, *args, **kwargs)

        if not _private_call(method, args):
            mapper = common.get_mapper(self)
            if mapper and mapper.auto_save:
                log.debug("Saving after call: %s", method.__name__)
                mapper.save()

        return result

    setattr(wrapped, '_save_after', True)

    return wrapped


def _private_call(method, args, prefix='_'):
    """Determine if a call's first argument is a private variable name."""
    if method.__name__ in ('__getattribute__', '__setattr__'):
        assert isinstance(args[0], str)
        return args[0].startswith(prefix)
    else:
        return False


class Mappable(metaclass=abc.ABCMeta):
    """Base class for objects with attributes mapped to file."""

    # pylint: disable=no-member

    @load_before
    def __getattribute__(self, name):
        return object.__getattribute__(self, name)

    @save_after
    def __setattr__(self, name, value):
        super().__setattr__(name, value)

    @load_before
    def __iter__(self):
        return super().__iter__()

    @load_before
    def __getitem__(self, key):
        return super().__getitem__(key)

    @save_after
    def __setitem__(self, key, value):
        super().__setitem__(key, value)

    @save_after
    def __delitem__(self, key):
        super().__delitem__(key)

    @save_after
    def append(self, *args, **kwargs):
        super().append(*args, **kwargs)

    @save_after
    def extend(self, *args, **kwargs):
        super().extend(*args, **kwargs)

    @save_after
    def insert(self, *args, **kwargs):
        super().insert(*args, **kwargs)

    @save_after
    def remove(self, *args, **kwargs):
        super().remove(*args, **kwargs)

    @save_after
    def pop(self, *args, **kwargs):
        super().pop(*args, **kwargs)

    @save_after
    def clear(self, *args, **kwargs):
        super().clear(*args, **kwargs)

    @save_after
    def sort(self, *args, **kwargs):
        super().sort(*args, **kwargs)

    @save_after
    def reverse(self, *args, **kwargs):
        super().reverse(*args, **kwargs)

    @save_after
    def popitem(self, *args, **kwargs):
        super().popitem(*args, **kwargs)

    @save_after
    def update(self, *args, **kwargs):
        super().update(*args, **kwargs)


_LOAD_BEFORE_METHODS = [
    '__getattribute__',
    '__iter__',
    '__getitem__',
]
_SAVE_AFTER_METHODS = [
    '__setattr__',
    '__setitem__',
    '__delitem__',
    'append',
    'extend',
    'insert',
    'remove',
    'pop',
    'clear',
    'sort',
    'reverse',
    'popitem',
    'update',
]


def patch_methods(instance):
    log.debug("Patching methods on: %r", instance)
    cls = instance.__class__

    for name in _LOAD_BEFORE_METHODS:
        try:
            method = getattr(cls, name)
        except AttributeError:
            log.trace("No method: %s", name)
        else:
            modified_method = load_before(method)
            setattr(cls, name, modified_method)
            log.trace("Patched to load before call: %s", name)

    for name in _SAVE_AFTER_METHODS:
        try:
            method = getattr(cls, name)
        except AttributeError:
            log.trace("No method: %s", name)
        else:
            modified_method = save_after(method)
            setattr(cls, name, modified_method)
            log.trace("Patched to save after call: %s", name)



"""Package for YORM."""

from . import bases, types
from .common import UUID
from .decorators import sync, sync_object, sync_instances, attr
from .utilities import create, find, match, load, save, delete
from .bases import Container, Converter, Mappable
from .mixins import ModelMixin

__project__ = 'YORM'
__version__ = '1.5.1'



import warnings

from yorm import utilities


class ModelMixin:
    """Adds ORM methods to a mapped class."""

    @classmethod
    def create(cls, *args, **kwargs):
        return utilities.create(cls, *args, **kwargs)

    @classmethod
    def new(cls, *args, **kwargs):
        msg = "ModelMixin.new() has been renamed to ModelMixin.create()"
        warnings.warn(msg, DeprecationWarning)
        return utilities.create(cls, *args, **kwargs)

    @classmethod
    def find(cls, *args, **kwargs):
        return utilities.find(cls, *args, **kwargs)

    @classmethod
    def match(cls, *args, **kwargs):
        return utilities.match(cls, *args, **kwargs)

    def load(self):
        return utilities.load(self)

    def save(self):
        return utilities.save(self)

    def delete(self):
        return utilities.delete(self)



"""Functions to enable mapping on classes and instances."""

import uuid
from collections import OrderedDict
import logging

from . import common
from .bases.mappable import patch_methods
from .mapper import Mapper

log = logging.getLogger(__name__)


def sync(*args, **kwargs):
    """Decorate class or map object based on arguments.

    This function will call either:

    * `sync_object` - when given an unmapped object
    * `sync_instances` - when used as the class decorator

    Consult the signature of each call for more information.

    """
    if 'path_format' in kwargs or args and isinstance(args[0], str):
        return sync_instances(*args, **kwargs)
    else:
        return sync_object(*args, **kwargs)


def sync_object(instance, path, attrs=None, **kwargs):
    """Enable YAML mapping on an object.

    :param instance: object to patch with YAML mapping behavior
    :param path: file path for dump/parse
    :param attrs: dictionary of attribute names mapped to converter classes

    :param auto_create: automatically create the file to save attributes
    :param auto_save: automatically save attribute changes to the file
    :param auto_track: automatically add new attributes from the file

    """
    log.info("Mapping %r to %s...", instance, path)

    common.get_mapper(instance, expected=False)
    patch_methods(instance)

    attrs = _ordered(attrs) or common.attrs[instance.__class__]
    mapper = Mapper(instance, path, attrs, **kwargs)

    if mapper.missing:
        if mapper.auto_create:
            mapper.create()
            if mapper.auto_save:
                mapper.save()
                mapper.load()
    else:
        mapper.load()

    common.set_mapper(instance, mapper)
    log.info("Mapped %r to %s", instance, path)

    return instance


def sync_instances(path_format, format_spec=None, attrs=None, **kwargs):
    """Decorate class to enable YAML mapping after instantiation.

    :param path_format: formatting string to create file paths for dump/parse
    :param format_spec: dictionary to use for string formatting
    :param attrs: dictionary of attribute names mapped to converter classes

    :param auto_create: automatically create the file to save attributes
    :param auto_save: automatically save attribute changes to the file
    :param auto_track: automatically add new attributes from the file

    """
    format_spec = format_spec or {}
    attrs = attrs or OrderedDict()

    def decorator(cls):
        """Class decorator to map instances to files."""
        common.path_formats[cls] = path_format
        init = cls.__init__

        def modified_init(self, *_args, **_kwargs):
            init(self, *_args, **_kwargs)

            log.info("Mapping instance of %r to '%s'...", cls, path_format)

            format_values = {}
            for key, value in format_spec.items():
                format_values[key] = getattr(self, value)
            if '{' + common.UUID + '}' in path_format:
                format_values[common.UUID] = uuid.uuid4().hex
            format_values['self'] = self

            common.attrs[cls].update(attrs)
            common.attrs[cls].update(common.attrs[self.__class__])
            path = path_format.format(**format_values)
            sync_object(self, path, **kwargs)

        modified_init.__doc__ = init.__doc__
        cls.__init__ = modified_init

        return cls

    return decorator


def attr(**kwargs):
    """Class decorator to map attributes to types.

    :param kwargs: keyword arguments mapping attribute name to converter class

    """
    if len(kwargs) != 1:
        raise ValueError("Single attribute required: {}".format(kwargs))

    def decorator(cls):
        """Class decorator."""
        previous = common.attrs[cls]
        common.attrs[cls] = OrderedDict()
        for name, converter in kwargs.items():
            common.attrs[cls][name] = converter
        for name, converter in previous.items():
            common.attrs[cls][name] = converter

        return cls

    return decorator


def _ordered(data):
    """Sort a dictionary-like object by key."""
    if data is None:
        return None
    return OrderedDict(sorted(data.items(), key=lambda pair: pair[0]))



"""Integration tests to ensure a mapped object's signature remains unchanged."""
# pylint: disable=redefined-outer-name,unused-variable,expression-not-assigned

import pytest
from expecter import expect

from yorm import decorators


class SampleWithMagicMethods:

    def __init__(self):
        self.values = []

    def __setattr__(self, name, value):
        if name == 'foobar':
            self.values.append(('__setattr__', name, value))
        super().__setattr__(name, value)


@pytest.yield_fixture
def mapped():
    import yorm
    yorm.settings.fake = True
    yield decorators.sync(SampleWithMagicMethods(), "sample.yml")
    yorm.settings.fake = False


@pytest.fixture
def unmapped():
    return SampleWithMagicMethods()


def fuzzy_repr(obj):
    return repr(obj).split(" object at ")[0] + ">"


def describe_mapped_object():

    def it_retains_representation(mapped, unmapped):
        expect(fuzzy_repr(mapped)) == fuzzy_repr(unmapped)

    def it_retains_docstring(mapped, unmapped):
        expect(mapped.__doc__) == unmapped.__doc__

    def it_retains_class(mapped, unmapped):
        expect(mapped.__class__) == unmapped.__class__

    def it_retains_magic_methods(mapped):
        setattr(mapped, 'foobar', 42)
        expect(mapped.values) == [('__setattr__', 'foobar', 42)]

    def it_does_not_affect_unmapped_objects(mapped, unmapped):
        expect(hasattr(mapped, '__mapper__')) is True
        expect(hasattr(unmapped, '__mapper__')) is False
        expect(hasattr(mapped.__setattr__, '_modified_by_yorm')) is True
        expect(hasattr(unmapped.__setattr__, '_modified_by_yorm')) is False



# pylint: disable=missing-docstring,no-self-use,no-member,misplaced-comparison-constant,expression-not-assigned

import logging
from unittest.mock import patch, Mock

import pytest
from expecter import expect

import yorm
from yorm import common
from yorm.decorators import attr
from yorm.types import Dictionary, List
from yorm.types import String, Integer

from . import strip

log = logging.getLogger(__name__)


# CLASSES ######################################################################


@attr(abc=Integer)
class SampleDictionary(Dictionary):
    """Sample dictionary container."""


@attr(var1=Integer)
@attr(var2=String)
class SampleDictionaryWithInitialization(Dictionary):
    """Sample dictionary container with initialization."""

    def __init__(self, var1, var2, var3):
        super().__init__()
        self.var1 = var1
        self.var2 = var2
        self.var3 = var3


@attr(all=String)
class StringList(List):
    """Sample list container."""


class UnknownList(List):
    """Sample list container."""


# TESTS ########################################################################


class TestDictionary:
    """Unit tests for the `Dictionary` container."""

    obj = {'abc': 123}

    class SampleClass:

        def __init__(self):
            self.abc = 42

    class SampleClass2:

        def __init__(self):
            self.unmapped = Mock()

    data_value = [
        (obj, obj),
        (None, {'abc': 0}),
        ("key=value", {'key': "value", 'abc': 0}),
        ("key=", {'key': "", 'abc': 0}),
        ("key", {'key': None, 'abc': 0}),
    ]

    value_data = [
        (obj, obj),
        (SampleClass(), {'abc': 42}),
        (SampleClass2(), {'abc': 0}),
        ([], {'abc': 0}),
    ]

    def setup_method(self, _):
        """Reset the class' mapped attributes before each test."""
        common.attrs[SampleDictionary] = {'abc': Integer}

    @pytest.mark.parametrize("data,value", data_value)
    def test_to_value(self, data, value):
        """Verify input data is converted to values."""
        assert value == SampleDictionary.to_value(data)

    @pytest.mark.parametrize("value,data", value_data)
    def test_to_data(self, value, data):
        """Verify values are converted to output data."""
        assert data == SampleDictionary.to_data(value)

    def test_not_implemented(self):
        """Verify `Dictionary` cannot be used directly."""
        with pytest.raises(NotImplementedError):
            Dictionary()

    def test_dict_as_object(self):
        """Verify a `Dictionary` can be used as an attribute."""
        dictionary = SampleDictionaryWithInitialization(1, 2, 3.0)
        value = {'var1': 1, 'var2': '2'}
        value2 = dictionary.to_value(dictionary)
        assert value == value2
        # keys are not accessible as attributes
        assert not hasattr(value2, 'var1')
        assert not hasattr(value2, 'var2')
        assert not hasattr(value2, 'var3')

    def test_unknown_attrributes_are_ignored(self):
        obj = SampleDictionary.create_default()
        obj.update_value({'key': "value", 'abc': 7}, auto_track=False)
        assert {'abc': 7} == obj


class TestList:
    """Unit tests for the `List` container."""

    obj = ["a", "b", "c"]

    data_value = [
        (obj, obj),
        (None, []),
        ([None], []),
        ("a b c", ["a", "b", "c"]),
        ("a,b,c", ["a", "b", "c"]),
        ("abc", ["abc"]),
        ("a\nb\nc", ["a", "b", "c"]),
        (4.2, ['4.2']),
        (("a", "b"), ["a", "b"]),
    ]

    value_data = [
        (obj, obj),
        ([], [None]),
    ]

    @pytest.mark.parametrize("data,value", data_value)
    def test_to_value(self, data, value):
        """Verify input data is converted to values."""
        assert value == StringList.to_value(data)

    @pytest.mark.parametrize("value,data", value_data)
    def test_to_data(self, value, data):
        """Verify values are converted to output data."""
        assert data == StringList.to_data(value)

    def test_item_type(self):
        """Verify list item type can be determined."""
        assert String == StringList.item_type

    def test_item_type_none(self):
        """Verify list item type defaults to None."""
        assert None is UnknownList.item_type

    def test_not_implemented(self):
        """Verify `List` cannot be used directly."""
        with pytest.raises(NotImplementedError):
            List()
        with pytest.raises(NotImplementedError):
            UnknownList()

    def test_shortened_syntax(self):
        cls = List.of_type(Integer)
        expect(cls.__name__) == "IntegerList"
        expect(common.attrs[cls]) == {'all': Integer}


class TestExtensions:
    """Unit tests for extensions to the container classes."""

    class FindMixin:

        def find(self, value):
            for value2 in self:
                if value.lower() == value2.lower():
                    return value2
            return None

    @yorm.attr(a=yorm.types.String)
    class MyDictionary(Dictionary, FindMixin):
        pass

    @yorm.attr(all=yorm.types.String)
    class MyList(List, FindMixin):
        pass

    def test_converted_dict_keeps_type(self):
        my_dict = self.MyDictionary()
        my_dict['a'] = 1
        my_dict2 = self.MyDictionary.to_value(my_dict)
        assert 'a' == my_dict2.find('A')
        assert None is my_dict2.find('B')

    def test_converted_list_keeps_type(self):
        my_list = self.MyList()
        my_list.append('a')
        my_list2 = self.MyList.to_value(my_list)
        assert 'a' == my_list2.find('A')
        assert None is my_list2.find('B')


@patch('yorm.settings.fake', True)
class TestReservedNames:

    class MyObject:

        def __init__(self, items=None):
            self.items = items or []

        def __repr__(self):
            return "<my_object>"

    def test_list_named_items(self):
        my_object = self.MyObject()
        yorm.sync_object(my_object, "fake/path", {'items': StringList})

        log.info("Appending value to list of items...")
        my_object.items.append('foo')

        log.info("Checking object contents...")
        assert strip("""
        items:
        - foo
        """) == my_object.__mapper__.text

        log.info("Writing new file contents...")
        my_object.__mapper__.text = strip("""
        items:
        - bar
        """)

        log.info("Checking file contents...")
        assert ['bar'] == my_object.items



"""Converter classes for extensions to builtin types."""

import re
import logging

from .standard import String, Integer, Float, Boolean
from .containers import Dictionary, List
from ._representers import LiteralString


log = logging.getLogger(__name__)


# NULLABLE BUILTINS ############################################################


class NullableString(String):
    """Converter for the `str` type with `None` as default."""

    DEFAULT = None


class NullableInteger(Integer):
    """Converter for the `int` type with `None` as default."""

    DEFAULT = None


class NullableFloat(Float):
    """Converter for the `float` type with `None` as default."""

    DEFAULT = None


class NullableBoolean(Boolean):
    """Converter for the `bool` type with `None` as default."""

    DEFAULT = None


# CUSTOM TYPES #################################################################


class Markdown(String):
    """Converter for a `str` type that contains Markdown."""

    REGEX_MARKDOWN_SPACES = re.compile(r"""

    ([^\n ])  # any character but a newline or space

    (\ ?\n)   # optional space + single newline

    (?!       # none of the following:

      (?:\s)       # whitespace
      |
      (?:[-+*]\s)  # unordered list separator + whitespace
      |
      (?:\d+\.\s)  # number + period + whitespace

    )

    ([^\n])  # any character but a newline

    """, re.VERBOSE | re.IGNORECASE)

    # based on: http://en.wikipedia.org/wiki/Sentence_boundary_disambiguation
    REGEX_SENTENCE_BOUNDARIES = re.compile(r"""

    (            # one of the following:

      (?<=[a-z)][.?!])      # lowercase letter + punctuation
      |
      (?<=[a-z0-9][.?!]\")  # lowercase letter/number + punctuation + quote

    )

    (\s)          # any whitespace

    (?=\"?[A-Z])  # optional quote + an upppercase letter

    """, re.VERBOSE)

    @classmethod
    def to_value(cls, obj):
        """Join non-meaningful line breaks."""
        value = String.to_value(obj)
        return cls._join(value)

    @classmethod
    def to_data(cls, obj):
        """Break a string at sentences and dump as a literal string."""
        value = String.to_value(obj)
        data = String.to_data(value)
        split = cls._split(data)
        return LiteralString(split)

    @classmethod
    def _join(cls, text):
        r"""Convert single newlines (ignored by Markdown) to spaces.

        >>> Markdown._join("abc\n123")
        'abc 123'

        >>> Markdown._join("abc\n\n123")
        'abc\n\n123'

        >>> Markdown._join("abc \n123")
        'abc 123'

        """
        return cls.REGEX_MARKDOWN_SPACES.sub(r'\1 \3', text).strip()

    @classmethod
    def _split(cls, text, end='\n'):
        r"""Replace sentence boundaries with newlines and append a newline.

        :param text: string to line break at sentences
        :param end: appended to the end of the update text

        >>> Markdown._split("Hello, world!", end='')
        'Hello, world!'

        >>> Markdown._split("Hello, world! How are you? I'm fine. Good.")
        "Hello, world!\nHow are you?\nI'm fine.\nGood.\n"

        """
        stripped = text.strip()
        if stripped:
            return cls.REGEX_SENTENCE_BOUNDARIES.sub('\n', stripped) + end
        else:
            return ''


# CUSTOM CONTAINERS ############################################################


class AttributeDictionary(Dictionary):
    """Dictionary converter with keys available as attributes."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__dict__ = self

    @classmethod
    def create_default(cls):
        """Create an uninitialized object with keys as attributes."""
        if cls is AttributeDictionary:
            msg = "AttributeDictionary class must be subclassed to use"
            raise NotImplementedError(msg)

        try:
            obj = cls()
        except TypeError as exc:
            log.info("Default values in %s are not available when "
                     "positional arguments are used: %s", cls.__name__, exc)
            obj = cls.__new__(cls)
            obj.__dict__ = obj

        return obj


class SortedList(List):
    """List converter that is sorted on disk."""

    @classmethod
    def create_default(cls):
        """Create an uninitialized object."""
        if cls is SortedList:
            msg = "SortedList class must be subclassed to use"
            raise NotImplementedError(msg)
        if not cls.item_type:
            msg = "SortedList subclass must specify item type"
            raise NotImplementedError(msg)

        return cls.__new__(cls)

    @classmethod
    def to_data(cls, obj):
        """Convert all attribute values for optimal dumping to YAML."""
        value = cls.to_value(obj)

        data = []

        for item in sorted(value):
            data.append(cls.item_type.to_data(item))  # pylint: disable=no-member

        return data



"""Core object-file mapping functionality."""

import functools
from pprint import pformat
import logging

from . import common, diskutils, exceptions, types, settings
from .bases import Container

log = logging.getLogger(__name__)


def file_required(method):
    """Decorate methods that require the file to exist."""

    @functools.wraps(method)
    def wrapped(self, *args, **kwargs):
        if self.deleted:
            msg = "File deleted: {}".format(self.path)
            raise exceptions.DeletedFileError(msg)
        elif self.missing and not settings.fake:
            msg = "File missing: {}".format(self.path)
            raise exceptions.MissingFileError(msg)
        else:
            return method(self, *args, **kwargs)

    return wrapped


def prevent_recursion(method):
    """Decorate methods to prevent indirect recursive calls."""

    @functools.wraps(method)
    def wrapped(self, *args, **kwargs):
        # pylint: disable=protected-access
        if self._activity:
            return
        self._activity = True
        result = method(self, *args, **kwargs)
        self._activity = False
        return result

    return wrapped


def prefix(obj):
    """Prefix a string with a fake designator if enabled."""
    fake = "(fake) " if settings.fake else ""
    name = obj if isinstance(obj, str) else "'{}'".format(obj)
    return fake + name


class Mapper:
    """Utility class to map an object's attributes to a file.

    To start mapping attributes to a file:

        create -> [empty] -> FILE

    When getting an attribute:

        FILE -> read -> [text] -> parse -> [dict] -> load -> ATTRIBUTES

    When setting an attribute:

        ATTRIBUTES -> save -> [dict] -> dump -> [text] -> write -> FILE

    After the mapped file is no longer needed:

        delete -> [null] -> FILE

    """

    def __init__(self, obj, path, attrs, *,
                 auto_create=True, auto_save=True,
                 auto_track=False, auto_resolve=False):
        self._obj = obj
        self.path = path
        self.attrs = attrs
        self.auto_create = auto_create
        self.auto_save = auto_save
        self.auto_track = auto_track
        self.auto_resolve = auto_resolve

        self.exists = diskutils.exists(self.path)
        self.deleted = False
        self.auto_save_after_load = False

        self._activity = False
        self._timestamp = 0
        self._fake = ""

    def __str__(self):
        return str(self.path)

    @property
    def missing(self):
        return not self.exists

    @property
    def modified(self):
        """Determine if the file has been modified."""
        if settings.fake:
            changes = self._timestamp is not None
            return changes
        elif not self.exists:
            return True
        else:
            # TODO: this raises an exception is the file is missing
            was = self._timestamp
            now = diskutils.stamp(self.path)
            return was != now

    @modified.setter
    @file_required
    def modified(self, changes):
        """Mark the file as modified if there are changes."""
        if changes:
            log.debug("Marked %s as modified", prefix(self))
            self._timestamp = 0
        else:
            if settings.fake or self.path is None:
                self._timestamp = None
            else:
                self._timestamp = diskutils.stamp(self.path)
            log.debug("Marked %s as unmodified", prefix(self))

    @property
    def text(self):
        """Get file contents as a string."""
        log.info("Getting contents of %s...", prefix(self))
        if settings.fake:
            text = self._fake
        else:
            text = self._read()
        log.trace("Text read: \n%s", text[:-1])
        return text

    @text.setter
    def text(self, text):
        """Set file contents from a string."""
        log.info("Setting contents of %s...", prefix(self))
        if settings.fake:
            self._fake = text
        else:
            self._write(text)
        log.trace("Text wrote: \n%s", text.rstrip())
        self.modified = True

    @property
    def data(self):
        """Get the file values as a dictionary."""
        text = self._read()
        try:
            data = diskutils.parse(text, self.path)
        except ValueError as e:
            if not self.auto_resolve:
                raise e from None

            log.debug(e)
            log.warning("Clearing invalid contents: %s", self.path)
            self._write("")
            return {}

        log.trace("Parsed data: \n%s", pformat(data))
        return data

    @data.setter
    def data(self, data):
        """Set the file values from a dictionary."""
        text = diskutils.dump(data, self.path)
        self._write(text)

    def create(self):
        """Create a new file for the object."""
        log.info("Creating %s for %r...", prefix(self), self._obj)
        if self.exists:
            log.warning("Already created: %s", self)
            return
        if not settings.fake:
            diskutils.touch(self.path)
        self.exists = True
        self.deleted = False

    @file_required
    @prevent_recursion
    def load(self):
        """Update the object's mapped attributes from its file."""
        log.info("Loading %r from %s...", self._obj, prefix(self))

        # Update all attributes
        attrs2 = self.attrs.copy()
        for name, data in self.data.items():
            attrs2.pop(name, None)

            # Find a matching converter
            try:
                converter = self.attrs[name]
            except KeyError:
                if self.auto_track:
                    converter = types.match(name, data)
                    self.attrs[name] = converter
                else:
                    msg = "Ignored unknown file attribute: %s = %r"
                    log.warning(msg, name, data)
                    continue

            # Convert the parsed value to the attribute's final type
            attr = getattr(self._obj, name, None)
            if isinstance(attr, converter) and \
                    issubclass(converter, Container):
                attr.update_value(data, auto_track=self.auto_track)
            else:
                log.trace("Converting attribute %r to %r", name, converter)
                attr = converter.to_value(data)
                setattr(self._obj, name, attr)
            self._remap(attr, self)
            log.trace("Value loaded: %s = %r", name, attr)

        # Add missing attributes
        for name, converter in attrs2.items():
            try:
                existing_attr = getattr(self._obj, name)
            except AttributeError:
                value = converter.create_default()
                msg = "Default value for missing object attribute: %s = %r"
                log.warning(msg, name, value)
                setattr(self._obj, name, value)
                self._remap(value, self)
            else:
                if issubclass(converter, Container) and \
                        not isinstance(existing_attr, converter):
                    msg = "Converting container attribute %r to %r"
                    log.trace(msg, name, converter)
                    value = converter.create_default()
                    setattr(self._obj, name, value)
                    self._remap(value, self)

        # Set meta attributes
        self.modified = False

    def _remap(self, obj, root):
        """Attach mapper on nested attributes."""
        if isinstance(obj, Container):
            common.set_mapper(obj, root)

            if isinstance(obj, dict):
                for obj2 in obj.values():
                    self._remap(obj2, root)
            else:
                assert isinstance(obj, list)
                for obj2 in obj:
                    self._remap(obj2, root)

    @file_required
    @prevent_recursion
    def save(self):
        """Format and save the object's mapped attributes to its file."""
        log.info("Saving %r to %s...", self._obj, prefix(self))

        # Format the data items
        data = self.attrs.__class__()
        for name, converter in self.attrs.items():
            try:
                value = getattr(self._obj, name)
            except AttributeError:
                data2 = converter.to_data(None)
                msg = "Default data for missing object attribute: %s = %r"
                log.warning(msg, name, data2)
            else:
                data2 = converter.to_data(value)

            log.trace("Data to save: %s = %r", name, data2)
            data[name] = data2

        # Save the formatted to disk
        self.data = data

        # Set meta attributes
        self.modified = True
        self.auto_save_after_load = self.auto_save

    def delete(self):
        """Delete the object's file from the file system."""
        if self.exists:
            log.info("Deleting %s...", prefix(self))
            diskutils.delete(self.path)
        else:
            log.warning("Already deleted: %s", self)
        self.exists = False
        self.deleted = True

    @file_required
    def _read(self):
        """Read text from the object's file."""
        if settings.fake:
            return self._fake
        elif not self.exists:
            return ""
        else:
            return diskutils.read(self.path)

    @file_required
    def _write(self, text):
        """Write text to the object's file."""
        if settings.fake:
            self._fake = text
        else:
            diskutils.write(text, self.path)



"""Integration tests for the `yorm.settings.fake` option."""
# pylint: disable=missing-docstring,no-self-use,no-member,misplaced-comparison-constant

import os
from unittest.mock import patch

import yorm

from . import strip


# CLASSES ######################################################################


@yorm.attr(value=yorm.types.standard.Integer)
@yorm.sync("tmp/path/to/{self.name}.yml")
class Sample:
    """Sample class for fake mapping."""

    def __init__(self, name):
        self.name = name
        self.value = 0

    def __repr__(self):
        return "<sample {}>".format(id(self))


# TESTS ########################################################################


@patch('yorm.settings.fake', True)
class TestFake:
    """Integration tests with `yorm.settings.fake` enabled."""

    def test_no_file_create_when_fake(self, tmpdir):
        tmpdir.chdir()
        sample = Sample('sample')

        # ensure no file is created
        assert "tmp/path/to/sample.yml" == sample.__mapper__.path
        assert not os.path.exists(sample.__mapper__.path)

        # change object values
        sample.value = 42

        # check fake file
        assert strip("""
        value: 42
        """) == sample.__mapper__.text

        # ensure no file is created
        assert not os.path.exists(sample.__mapper__.path)

        # change fake file
        sample.__mapper__.text = "value: 0\n"

        # check object values
        assert 0 == sample.value

        # ensure no file is created
        assert not os.path.exists(sample.__mapper__.path)

    def test_fake_changes_indicate_modified(self, tmpdir):
        tmpdir.chdir()
        sample = Sample('sample')

        assert False is sample.__mapper__.modified
        assert 0 == sample.value

        sample.__mapper__.text = "value: 42\n"

        assert True is sample.__mapper__.modified
        assert 42 == sample.value
        assert False is sample.__mapper__.modified


class TestReal:
    """Integration tests with `yorm.settings.fake` disabled."""

    def test_setting_text_updates_attributes(self, tmpdir):
        tmpdir.chdir()
        sample = Sample('sample')

        sample.__mapper__.text = "value: 42"

        assert 42 == sample.value

    def test_setting_attributes_update_text(self, tmpdir):
        tmpdir.chdir()
        sample = Sample('sample')

        sample.value = 42

        assert strip("""
        value: 42
        """) == sample.__mapper__.text



"""Integration tests using YORM as a persistence model."""
# pylint: disable=missing-docstring,no-self-use,misplaced-comparison-constant

import os

from expecter import expect

import yorm
from yorm.types import String


# CLASSES ######################################################################


class Config:
    """Domain model."""

    def __init__(self, key, name=None, root=None):
        self.key = key
        self.name = name or ""
        self.root = root or ""


@yorm.attr(key=String)
@yorm.attr(name=String)
@yorm.sync("{self.root}/{self.key}/config.yml",
           auto_create=False, auto_save=False)
class ConfigModel:
    """Persistence model."""

    def __init__(self, key, root):
        self.key = key
        self.root = root
        print(self.key)
        self.unmapped = 0

    @staticmethod
    def pm_to_dm(model):
        config = Config(model.key)
        config.name = model.name
        config.root = model.root
        return config


class ConfigStore:

    def __init__(self, root):
        self.root = root

    def read(self, key):
        return yorm.find(ConfigModel, self.root, key)


# TESTS ########################################################################


class TestPersistanceMapping:  # pylint: disable=no-member

    root = os.path.join(os.path.dirname(__file__), 'files')

    def test_load_pm(self):
        model = ConfigModel('my_key', self.root)

        print(model.__dict__)
        assert model.key == "my_key"
        assert model.root == self.root
        assert model.name == "my_name"

    def test_create_dm_from_pm(self):
        model = ConfigModel('my_key', self.root)
        config = ConfigModel.pm_to_dm(model)

        print(config.__dict__)
        assert config.key == "my_key"
        assert config.root == self.root
        assert config.name == "my_name"

    def test_nonmapped_attribute_is_kept(self):
        model = ConfigModel('my_key', self.root)
        model.unmapped = 42
        assert 42 == model.unmapped

    def test_missing_files_are_handled(self):
        model = ConfigModel('my_key_manual', self.root)

        with expect.raises(yorm.exceptions.MissingFileError):
            print(model.name)


class TestStore:

    def test_read_missing(self, tmpdir):
        store = ConfigStore(str(tmpdir))
        assert None is store.read('unknown')



# pylint: disable=redefined-outer-name,expression-not-assigned,attribute-defined-outside-init,no-member

from expecter import expect

import yorm

from . import strip


@yorm.attr(status=yorm.types.Boolean)
@yorm.attr(label=yorm.types.String)
class StatusDictionary(yorm.types.Dictionary):
    """Sample dictionary converter with ordered attributes."""


@yorm.attr(string=yorm.types.String)
@yorm.attr(number_int=yorm.types.Integer)
@yorm.attr(dictionary=StatusDictionary)
@yorm.attr(number_real=yorm.types.Float)
@yorm.attr(truthy=yorm.types.Boolean)
@yorm.attr(falsey=yorm.types.Boolean)
@yorm.sync("sample.yml")
class Sample:
    """Sample class with ordered attributes."""


def test_attribute_order_is_maintained(tmpdir):
    tmpdir.chdir()
    sample = Sample()
    sample.string = "Hello, world!"
    sample.number_int = 42
    sample.number_real = 4.2
    # pylint: disable=duplicate-code
    sample.truthy = False
    sample.falsey = True
    sample.dictionary['status'] = 1

    expect(sample.__mapper__.text) == strip("""
    string: Hello, world!
    number_int: 42
    dictionary:
      status: true
      label: ''
    number_real: 4.2
    truthy: false
    falsey: true
    """)


def test_existing_files_are_reorderd(tmpdir):
    tmpdir.chdir()
    with open("sample.yml", 'w') as stream:
        stream.write(strip("""
        falsey: 1
        number_int: 2
        number_real: 3
        string: 4
        truthy: 5
        dictionary: {label: foo}
        """))
    sample = Sample()
    sample.falsey = 0

    expect(sample.__mapper__.text) == strip("""
    string: 4
    number_int: 2
    dictionary:
      status: false
      label: foo
    number_real: 3.0
    truthy: true
    falsey: false
    """)



# pylint: disable=missing-docstring,redefined-outer-name,unused-variable,expression-not-assigned

import pytest
from expecter import expect

import yorm
from yorm import exceptions
from yorm.mapper import Mapper
from yorm.types import Integer


class MyObject:
    var1 = 1


@pytest.fixture
def obj():
    return MyObject()


@pytest.fixture
def attrs():
    return {'var2': Integer, 'var3': Integer}


@pytest.yield_fixture(params=("real", "fake"))
def mapper(tmpdir, obj, attrs, request):
    path = request.param + "/path/to/file"
    backup = yorm.settings.fake
    if "fake" in path:
        yorm.settings.fake = True
    elif "real" in path:
        tmpdir.chdir()
    yield Mapper(obj, path, attrs, auto_track=False)
    yorm.settings.fake = backup


@pytest.fixture
def mapper_real(tmpdir, obj, attrs):
    tmpdir.chdir()
    return Mapper(obj, "real/path/to/file", attrs)


@pytest.yield_fixture
def mapper_fake(obj, attrs):
    backup = yorm.settings.fake
    yorm.settings.fake = True
    yield Mapper(obj, "fake/path/to/file", attrs)
    yorm.settings.fake = backup


def describe_mapper():

    def describe_create():

        def it_creates_the_file(mapper_real):
            mapper_real.create()

            expect(mapper_real.path).exists()
            expect(mapper_real.exists).is_true()
            expect(mapper_real.deleted).is_false()

        def it_pretends_when_fake(mapper_fake):
            mapper_fake.create()

            expect(mapper_fake.path).missing()
            expect(mapper_fake.exists).is_true()
            expect(mapper_fake.deleted).is_false()

        def it_can_be_called_twice(mapper_real):
            mapper_real.create()
            mapper_real.create()  # should be ignored

            expect(mapper_real.path).exists()

    def describe_delete():

        def it_deletes_the_file(mapper):
            mapper.create()
            mapper.delete()

            expect(mapper.path).missing()
            expect(mapper.exists).is_false()
            expect(mapper.deleted).is_true()

        def it_can_be_called_twice(mapper):
            mapper.delete()
            mapper.delete()  # should be ignored

            expect(mapper.path).missing()

    def describe_load():

        def it_adds_missing_attributes(obj, mapper):
            mapper.create()
            mapper.load()

            expect(obj.var1) == 1
            expect(obj.var2) == 0
            expect(obj.var3) == 0

        def it_ignores_new_attributes(obj, mapper):
            mapper.create()
            mapper.text = "var4: foo"

            mapper.load()
            with expect.raises(AttributeError):
                print(obj.var4)

        def it_infers_types_on_new_attributes_with_auto_track(obj, mapper):
            mapper.auto_track = True
            mapper.create()
            mapper.text = "var4: foo"

            mapper.load()
            expect(obj.var4) == "foo"

            obj.var4 = 42
            mapper.save()

            mapper.load()
            expect(obj.var4) == "42"

        def it_raises_an_exception_after_delete(mapper):
            mapper.delete()

            with expect.raises(exceptions.DeletedFileError):
                mapper.load()

    def describe_modified():

        def is_true_initially(mapper):
            expect(mapper.modified).is_true()

        def is_true_after_create(mapper):
            mapper.create()

            expect(mapper.modified).is_true()

        def is_true_after_delete(mapper):
            mapper.delete()

            expect(mapper.modified).is_true()

        def is_false_after_load(mapper):
            mapper.create()
            mapper.load()

            expect(mapper.modified).is_false()

        def can_be_set_false(mapper):
            mapper.create()
            mapper.modified = False

            expect(mapper.modified).is_false()

        def can_be_set_true(mapper):
            mapper.create()
            mapper.modified = True

            expect(mapper.modified).is_true()

    def describe_text():

        def can_get_the_file_contents(obj, mapper):
            mapper.create()
            obj.var3 = 42
            mapper.save()

            expect(mapper.text) == "var2: 0\nvar3: 42\n"

        def can_set_the_file_contents(obj, mapper):
            mapper.create()
            mapper.text = "var2: 42\n"
            mapper.load()

            expect(obj.var2) == 42

    def describe_data():

        def can_get_the_file_values(obj, mapper):
            mapper.create()
            obj.var3 = 42
            mapper.save()

            expect(mapper.data) == {'var2': 0, 'var3': 42}

        def can_set_the_file_values(obj, mapper):
            mapper.create()
            mapper.data = {'var2': 42}
            mapper.load()

            expect(obj.var2) == 42

        def handles_invalid_content_if_enabled(mapper):
            mapper.auto_resolve = True
            mapper.create()
            mapper.text = "abc"

            expect(mapper.data) == {}



# pylint: disable=unused-variable,expression-not-assigned

from unittest.mock import patch, call

import pytest
from expecter import expect

import yorm
from yorm.mixins import ModelMixin


def describe_model_mixin():

    @pytest.fixture
    def mixed_class():

        @yorm.sync("tmp/model.yml")
        class MyClass(ModelMixin):
            pass

        return MyClass

    @pytest.fixture
    def mixed_instance(mixed_class):
        return mixed_class()

    @patch('yorm.mixins.utilities')
    def it_adds_a_create_method(utilities, mixed_class):
        mixed_class.create('foobar', overwrite=True)

        expect(utilities.mock_calls) == [
            call.create(mixed_class, 'foobar', overwrite=True)
        ]

    @patch('yorm.mixins.utilities')
    def it_adds_a_find_method(utilities, mixed_class):
        mixed_class.find('foobar', create=True)

        expect(utilities.mock_calls) == [
            call.find(mixed_class, 'foobar', create=True)
        ]

    @patch('yorm.mixins.utilities')
    def it_adds_a_match_method(utilities, mixed_class):
        mixed_class.match(foo='bar')

        expect(utilities.mock_calls) == [
            call.match(mixed_class, foo='bar')
        ]

    @patch('yorm.mixins.utilities')
    def it_adds_a_load_method(utilities, mixed_instance):
        mixed_instance.load()

        expect(utilities.mock_calls) == [
            call.load(mixed_instance)
        ]

    @patch('yorm.mixins.utilities')
    def it_adds_a_save_method(utilities, mixed_instance):
        mixed_instance.save()

        expect(utilities.mock_calls) == [
            call.save(mixed_instance)
        ]

    @patch('yorm.mixins.utilities')
    def it_adds_a_delete_method(utilities, mixed_instance):
        mixed_instance.delete()

        expect(utilities.mock_calls) == [
            call.delete(mixed_instance)
        ]



"""Functions to interact with mapped classes and instances."""

import inspect
import logging
import string
import glob
import types

import parse

from . import common, exceptions

log = logging.getLogger(__name__)


def create(class_or_instance, *args, overwrite=False, **kwargs):
    """Create a new mapped object.

    NOTE: Calling this function is unnecessary with 'auto_create' enabled.

    """
    instance = _instantiate(class_or_instance, *args, **kwargs)
    mapper = common.get_mapper(instance, expected=True)

    if mapper.exists and not overwrite:
        msg = "{!r} already exists".format(mapper.path)
        raise exceptions.DuplicateMappingError(msg)

    return load(save(instance))


def find(class_or_instance, *args, create=False, **kwargs):  # pylint: disable=redefined-outer-name
    """Find a matching mapped object or return None."""
    instance = _instantiate(class_or_instance, *args, **kwargs)
    mapper = common.get_mapper(instance, expected=True)

    if mapper.exists:
        return instance
    elif create:
        return save(instance)
    else:
        return None


class GlobFormatter(string.Formatter):
    """Uses '*' for all unknown fields."""

    WILDCARD = object()

    def get_field(self, field_name, args, kwargs):
        try:
            return super().get_field(field_name, args, kwargs)
        except (KeyError, IndexError, AttributeError):
            return self.WILDCARD, None

    def get_value(self, key, args, kwargs):
        try:
            return super().get_value(key, args, kwargs)
        except (KeyError, IndexError, AttributeError):
            return self.WILDCARD

    def convert_field(self, value, conversion):
        if value is self.WILDCARD:
            return self.WILDCARD
        else:
            return super().convert_field(value, conversion)

    def format_field(self, value, format_spec):
        if value is self.WILDCARD:
            return '*'
        else:
            return super().format_field(value, format_spec)


def _unpack_parsed_fields(pathfields):
    return {
        (k[len('self.'):] if k.startswith('self.') else k): v
        for k, v in pathfields.items()
    }


def match(cls_or_path, _factory=None, **kwargs):
    """Yield all matching mapped objects.

    Can be used two ways:

    * With a YORM-decorated class, optionally with a factory callable
    * With a Python 3-style string template and a factory callable

    The factory callable must accept keyword arguments, extracted from the file
    name merged with those passed to match(). If no factory is given, the class
    itself is used as the factory (same signature).

    Keyword arguments are used to filter objects. Filtering is only done by
    filename, so only fields that are part of the path_format can be filtered
    against.

    """
    if isinstance(cls_or_path, type):
        path_format = common.path_formats[cls_or_path]
        # Let KeyError fail through
        if _factory is None:
            _factory = cls_or_path
    else:
        path_format = cls_or_path
        if _factory is None:
            raise TypeError("Factory must be given if a path format is given")

    gf = GlobFormatter()
    mock = types.SimpleNamespace(**kwargs)

    kwargs['self'] = mock
    posix_pattern = gf.vformat(path_format, (), kwargs.copy())
    del kwargs['self']
    py_pattern = parse.compile(path_format)

    for filename in glob.iglob(posix_pattern):
        pathfields = py_pattern.parse(filename).named
        fields = _unpack_parsed_fields(pathfields)
        fields.update(kwargs)
        yield _factory(**fields)


def load(instance):
    """Force the loading of a mapped object's file.

    NOTE: Calling this function is unnecessary. It exists for the
        aesthetic purpose of having symmetry between save and load.

    """
    mapper = common.get_mapper(instance, expected=True)

    mapper.load()

    return instance


def save(instance):
    """Save a mapped object to file.

    NOTE: Calling this function is unnecessary with 'auto_save' enabled.

    """
    mapper = common.get_mapper(instance, expected=True)

    if mapper.deleted:
        msg = "{!r} was deleted".format(mapper.path)
        raise exceptions.DeletedFileError(msg)

    if not mapper.exists:
        mapper.create()

    mapper.save()

    return instance


def delete(instance):
    """Delete a mapped object's file."""
    mapper = common.get_mapper(instance, expected=True)

    mapper.delete()

    return None


def _instantiate(class_or_instance, *args, **kwargs):
    if inspect.isclass(class_or_instance):
        instance = class_or_instance(*args, **kwargs)
    else:
        assert not args
        instance = class_or_instance

    return instance



# pylint: disable=missing-docstring,no-self-use,attribute-defined-outside-init,protected-access,misplaced-comparison-constant

import logging
from unittest.mock import Mock

import pytest

import yorm
from yorm.bases import Mappable
from yorm.mapper import Mapper
from yorm.types import String, Integer, Boolean, List, Dictionary

from . import strip


class MockMapper(Mapper):
    """Mapped file with stubbed file IO."""

    def __init__(self, obj, path, attrs):
        super().__init__(obj, path, attrs)
        self._mock_file = None
        self._mock_modified = True
        self.exists = True

    def _read(self):
        text = self._mock_file
        logging.debug("Mock read:\n%s", text.strip())
        return text

    def _write(self, text):
        logging.debug("Mock write:\n%s", text.strip())
        self._mock_file = text
        self.modified = True

    @property
    def modified(self):
        return self._mock_modified

    @modified.setter
    def modified(self, changes):
        self._mock_modified = changes


# CLASSES ######################################################################

@yorm.attr(all=Integer)
class IntegerList(List):
    """List of integers."""


@yorm.attr(status=Boolean)
class StatusDictionary(Dictionary):
    """Dictionary of statuses."""


class SampleMappable(Mappable):
    """Sample mappable class with hard-coded settings."""

    def __init__(self):
        self.__mapper__ = None

        logging.debug("Initializing sample...")
        self.var1 = None
        self.var2 = None
        self.var3 = None
        self.var4 = None
        self.var5 = None
        logging.debug("Sample initialized")

        path = "mock/path/to/sample.yml"
        attrs = {'var1': String,
                 'var2': Integer,
                 'var3': Boolean,
                 'var4': IntegerList,
                 'var5': StatusDictionary}
        self.__mapper__ = MockMapper(self, path, attrs)
        self.__mapper__.save()

    def __repr__(self):
        return "<sample {}>".format(id(self))


# TESTS ########################################################################


class TestMappable:
    """Unit tests for the `Mappable` class."""

    def setup_method(self, _):
        """Create an mappable instance for tests."""
        self.sample = SampleMappable()

    def test_init(self):
        """Verify files are created after initialized."""
        text = self.sample.__mapper__._read()
        assert strip("""
        var1: ''
        var2: 0
        var3: false
        var4:
        -
        var5:
          status: false
        """) == text

    def test_set(self):
        """Verify the file is written to after setting an attribute."""
        self.sample.var1 = "abc123"
        self.sample.var2 = 1
        self.sample.var3 = True
        self.sample.var4 = [42]
        self.sample.var5 = {'status': True}
        text = self.sample.__mapper__._read()
        assert strip("""
        var1: abc123
        var2: 1
        var3: true
        var4:
        - 42
        var5:
          status: true
        """) == text

    def test_set_converted(self):
        """Verify conversion occurs when setting attributes."""
        self.sample.var1 = 42
        self.sample.var2 = "1"
        self.sample.var3 = 'off'
        self.sample.var4 = None
        self.sample.var5 = {'status': 1}
        text = self.sample.__mapper__._read()
        assert strip("""
        var1: 42
        var2: 1
        var3: false
        var4:
        -
        var5:
          status: true
        """) == text

    def test_set_error(self):
        """Verify an exception is raised when a value cannot be converted."""
        with pytest.raises(ValueError):
            self.sample.var2 = "abc"

    def test_get(self):
        """Verify the file is read from before getting an attribute."""
        text = strip("""
        var1: def456
        var2: 42
        var3: off
        """)
        self.sample.__mapper__._write(text)
        assert"def456" == self.sample.var1
        assert 42 == self.sample.var2
        assert False is self.sample.var3

    def test_error_invalid_yaml(self):
        """Verify an exception is raised on invalid YAML."""
        text = strip("""
        invalid: -
        """)
        self.sample.__mapper__._write(text)
        with pytest.raises(ValueError):
            print(self.sample.var1)

    def test_error_unexpected_yaml(self):
        """Verify an exception is raised on unexpected YAML."""
        text = strip("""
        not a dictionary
        """)
        self.sample.__mapper__._write(text)
        with pytest.raises(ValueError):
            print(self.sample.var1)

    def test_new(self):
        """Verify new attributes are added to the object."""
        self.sample.__mapper__.auto_track = True
        text = strip("""
        new: 42
        """)
        self.sample.__mapper__._write(text)
        assert 42 == self.sample.new

    def test_new_unknown(self):
        """Verify an exception is raised on new attributes w/ unknown types"""
        self.sample.__mapper__.auto_track = True
        text = strip("""
        new: !!timestamp 2001-12-15T02:59:43.1Z
        """)
        self.sample.__mapper__._write(text)
        with pytest.raises(ValueError):
            print(self.sample.var1)


class TestMappableTriggers:

    class MockDict(Mappable, dict):
        pass

    class MockList:

        def append(self, value):
            print(value)

        def insert(self, value):
            print(value)

    class Sample(MockDict, MockList):

        __mapper__ = Mock()
        __mapper__.attrs = {}
        __mapper__.load = Mock()
        __mapper__.save = Mock()

    def setup_method(self, _):
        """Create an mappable instance for tests."""
        self.sample = self.Sample()
        self.sample.__mapper__.load.reset_mock()
        self.sample.__mapper__.save.reset_mock()
        self.sample.__mapper__.auto_save_after_load = False

    def test_getattribute(self):
        with pytest.raises(AttributeError):
            getattr(self.sample, 'foo')
        assert 1 == self.sample.__mapper__.load.call_count
        assert 0 == self.sample.__mapper__.save.call_count

    def test_setattr(self):
        self.sample.__mapper__.attrs['foo'] = Mock()
        setattr(self.sample, 'foo', 'bar')
        assert 0 == self.sample.__mapper__.load.call_count
        assert 1 == self.sample.__mapper__.save.call_count

    def test_getitem(self):
        with pytest.raises(KeyError):
            print(self.sample['foo'])
        assert 1 == self.sample.__mapper__.load.call_count
        assert 0 == self.sample.__mapper__.save.call_count

    def test_setitem(self):
        self.sample['foo'] = 'bar'
        assert 0 == self.sample.__mapper__.load.call_count
        assert 1 == self.sample.__mapper__.save.call_count

    def test_delitem(self):
        self.sample['foo'] = 'bar'
        self.sample.__mapper__.save.reset_mock()

        del self.sample['foo']
        assert 0 == self.sample.__mapper__.load.call_count
        assert 1 == self.sample.__mapper__.save.call_count

    def test_append(self):
        self.sample.append('foo')
        assert 1 == self.sample.__mapper__.load.call_count
        assert 1 == self.sample.__mapper__.save.call_count

    def test_insert(self):
        self.sample.insert('foo')
        assert 1 == self.sample.__mapper__.load.call_count
        assert 1 == self.sample.__mapper__.save.call_count

    def test_iter(self):
        self.sample.append('foo')
        self.sample.append('bar')
        self.sample.__mapper__.load.reset_mock()
        self.sample.__mapper__.save.reset_mock()
        self.sample.__mapper__.auto_save_after_load = False
        self.sample.__mapper__.modified = True

        for item in self.sample:
            print(item)
        assert 1 == self.sample.__mapper__.load.call_count
        assert 0 == self.sample.__mapper__.save.call_count

    def test_handle_missing_mapper(self):
        sample = self.MockDict()
        sample.__mapper__ = None
        sample[0] = 0
        print(sample[0])
        assert None is sample.__mapper__
